@@ -22,10 +22,14 @@
 #include <string.h>
 #include <stdio.h>
 #include <errno.h>
-#include <unistd.h>
+#ifdef HAVE_UNISTD_H
+# include <unistd.h>
+#endif
 #include <assert.h>
 #ifdef HAVE_W32_SYSTEM
-#include <process.h>
+#ifndef HAVE_W32CE_SYSTEM
+# include <process.h>
+#endif
 #endif
 #include "assuan-defs.h"
 
